package com.vladmihalcea.hibernate.type.array;

import com.vladmihalcea.hibernate.util.AbstractPostgreSQLIntegrationTest;
import com.vladmihalcea.hibernate.util.transaction.JPATransactionFunction;
import org.hibernate.annotations.Type;
import org.hibernate.annotations.TypeDef;
import org.hibernate.annotations.TypeDefs;
import org.hibernate.jpa.TypedParameterValue;
import org.junit.Before;
import org.junit.Test;

import javax.persistence.*;
import javax.sql.DataSource;
import java.sql.Connection;
import java.sql.SQLException;
import java.sql.Statement;

import static org.junit.Assert.assertArrayEquals;
import static org.junit.Assert.fail;

/**
 * @author Nikita Konev
 */
public class PostgreSQLEnumArrayTypeTest extends AbstractPostgreSQLIntegrationTest {

    public static final EnumArrayType ROLE_TYPE = new EnumArrayType(UserRole.class, "user_role");

    @Override
    protected Class<?>[] entities() {
        return new Class<?>[]{
                UserAccount.class
        };
    }

    @Before
    public void init() {
        DataSource dataSource = newDataSource();
        Connection connection = null;
        try {
            connection = dataSource.getConnection();
            Statement statement = null;
            try {
                statement = connection.createStatement();
                statement.executeUpdate("DROP TYPE IF EXISTS user_role;");
                statement.executeUpdate("CREATE TYPE user_role AS ENUM ('ROLE_ADMIN', 'ROLE_USER');");
            } finally {
                if (statement!=null){
                    statement.close();
                }
            }
        } catch (SQLException e) {
            fail(e.getMessage());
        } finally {
            if (connection!=null){
                try {
                    connection.close();
                } catch (SQLException e) {
                    fail(e.getMessage());
                }
            }
        }
        super.init();
    }

    @Test
    public void test() {
        final UserRole[] userRoles = {UserRole.ROLE_ADMIN, UserRole.ROLE_USER};

        doInJPA(new JPATransactionFunction<Void>() {
            @Override
            public Void apply(EntityManager entityManager) {
                UserAccount account = new UserAccount();
                account.setUsername("vladmihalcea.com");
                account.setRoles(userRoles);
                entityManager.persist(account);
                return null;
            }
        });

        doInJPA(new JPATransactionFunction<Void>() {
            @Override
            public Void apply(EntityManager entityManager) {
                UserAccount singleResult = entityManager
                .createQuery(
                    "select ua " +
                    "from UserAccountEntity ua " +
                    "where ua.username = :username", UserAccount.class)
                .setParameter("username", "vladmihalcea.com")
                .getSingleResult();

                assertArrayEquals(userRoles, singleResult.getRoles());
                return null;
            }
        });
    }

    @Test
    public void testTypedParameterValue() {
        final UserRole[] userRoles = {UserRole.ROLE_ADMIN, UserRole.ROLE_USER};
        final UserRole[] requiredRoles = {UserRole.ROLE_USER};

        doInJPA(new JPATransactionFunction<Void>() {
            @Override
            public Void apply(EntityManager entityManager) {
                UserAccount account = new UserAccount();
                account.setUsername("vladmihalcea.com");
                account.setRoles(userRoles);
                entityManager.persist(account);
                return null;
            }
        });

        doInJPA(new JPATransactionFunction<Void>() {
            @Override
            public Void apply(EntityManager entityManager) {
                entityManager
                        .createQuery(
                                "select ua " +
                                        "from UserAccountEntity ua " +
                                        "where ua.roles = :roles", UserAccount.class)
                        .setParameter("roles", new TypedParameterValue(ROLE_TYPE, requiredRoles))
                        .getResultList();
                return null;
            }
        });
    }

    public enum UserRole {
        ROLE_ADMIN,
        ROLE_USER,
    }

    @TypeDefs({
            @TypeDef(
                    name = "pgsql_array",
                    typeClass = EnumArrayType.class
            )

    })
    @Entity(name = "UserAccountEntity")
    @Table(name = "users")
    public static class UserAccount {

        @Id
        @GeneratedValue
        private Long id;

        private String username;

        @Type(
            type = "pgsql_array",
            parameters = @org.hibernate.annotations.Parameter(
                name = "sql_array_type",
                value = "user_role"
            )
        )
        @Column(
                name = "roles",
                columnDefinition = "user_role[]"
        )
        private UserRole roles[];

        public Long getId() {
            return id;
        }

        public void setId(Long id) {
            this.id = id;
        }

        public String getUsername() {
            return username;
        }

        public void setUsername(String username) {
            this.username = username;
        }

        public UserRole[] getRoles() {
            return roles;
        }

        public void setRoles(UserRole[] roles) {
            this.roles = roles;
        }
    }
}
